@@ -25,3 +25,5 @@ L3 = 'L3_ROUTER_NAT'
 LOG_API = "LOGGING"
 PORTFORWARDING = "PORTFORWARDING"
 FLOATINGIPPOOL = "FLOATINGIPPOOL"
+NETWORK_SEGMENT_RANGE = "NETWORK_SEGMENT_RANGE"
+CONNTRACKHELPER = "CONNTRACKHELPER"
