@@ -17,19 +17,42 @@ import mock
 from oslo_db import exception as db_exc
 from oslotest import base
 
-from neutron_lib._callbacks import events
-from neutron_lib._callbacks import exceptions
-from neutron_lib._callbacks import manager
-from neutron_lib._callbacks import resources
+from neutron_lib.callbacks import events
+from neutron_lib.callbacks import exceptions
+from neutron_lib.callbacks import manager
+from neutron_lib.callbacks import priority_group
+from neutron_lib.callbacks import resources
+
+
+PRI_HIGH = 0
+PRI_MED = 5000
+PRI_LOW = 10000
+
+
+class ObjectWithCallback(object):
+
+    def __init__(self):
+        self.counter = 0
+
+    def callback(self, *args, **kwargs):
+        self.counter += 1
+
+
+class GloriousObjectWithCallback(ObjectWithCallback):
+    pass
 
 
 def callback_1(*args, **kwargs):
     callback_1.counter += 1
+
+
 callback_id_1 = manager._get_id(callback_1)
 
 
 def callback_2(*args, **kwargs):
     callback_2.counter += 1
+
+
 callback_id_2 = manager._get_id(callback_2)
 
 
@@ -41,6 +64,10 @@ def callback_raise_retriable(*args, **kwargs):
     raise db_exc.DBDeadlock()
 
 
+def callback_3(resource, event, trigger, payload):
+    callback_3.counter += 1
+
+
 class CallBacksManagerTestCase(base.BaseTestCase):
 
     def setUp(self):
@@ -48,6 +75,7 @@ class CallBacksManagerTestCase(base.BaseTestCase):
         self.manager = manager.CallbacksManager()
         callback_1.counter = 0
         callback_2.counter = 0
+        callback_3.counter = 0
 
     def test_subscribe(self):
         self.manager.subscribe(
@@ -55,7 +83,8 @@ class CallBacksManagerTestCase(base.BaseTestCase):
         self.assertIsNotNone(
             self.manager._callbacks[resources.PORT][events.BEFORE_CREATE])
         self.assertIn(callback_id_1, self.manager._index)
-        self.assertEqual(self.__module__ + '.callback_1', callback_id_1)
+        self.assertEqual(self.__module__ + '.callback_1-%s' %
+                         hash(callback_1), callback_id_1)
 
     def test_subscribe_unknown(self):
         self.manager.subscribe(
@@ -82,11 +111,17 @@ class CallBacksManagerTestCase(base.BaseTestCase):
             callback_2, resources.PORT, events.BEFORE_CREATE)
         self.assertEqual(2, len(self.manager._index))
         self.assertEqual(
-            2,
+            1,
             len(self.manager._callbacks[resources.PORT][events.BEFORE_CREATE]))
+        self.assertEqual(
+            2,
+            len(self.manager._callbacks
+                [resources.PORT][events.BEFORE_CREATE][0][1]))
 
     def test_unsubscribe_during_iteration(self):
-        unsub = lambda r, e, *a, **k: self.manager.unsubscribe(unsub, r, e)
+        def unsub(r, e, *a, **k):
+            return self.manager.unsubscribe(unsub, r, e)
+
         self.manager.subscribe(unsub, resources.PORT,
                                events.BEFORE_CREATE)
         self.manager.notify(resources.PORT, events.BEFORE_CREATE, mock.ANY)
@@ -138,12 +173,17 @@ class CallBacksManagerTestCase(base.BaseTestCase):
         self.manager.subscribe(
             callback_2, resources.PORT, events.BEFORE_DELETE)
         self.manager.unsubscribe_by_resource(callback_1, resources.PORT)
-        self.assertNotIn(
-            callback_id_1,
-            self.manager._callbacks[resources.PORT][events.BEFORE_CREATE])
+        self.assertEqual(
+            0,
+            len(self.manager._callbacks
+                [resources.PORT][events.BEFORE_CREATE]))
+        self.assertEqual(
+            1,
+            len(self.manager._callbacks[resources.PORT][events.BEFORE_DELETE]))
         self.assertIn(
             callback_id_2,
-            self.manager._callbacks[resources.PORT][events.BEFORE_DELETE])
+            self.manager._callbacks
+            [resources.PORT][events.BEFORE_DELETE][0][1])
         self.assertNotIn(callback_id_1, self.manager._index)
 
     def test_unsubscribe_all(self):
@@ -246,3 +286,104 @@ class CallBacksManagerTestCase(base.BaseTestCase):
         self.manager.clear()
         self.assertEqual(0, len(self.manager._callbacks))
         self.assertEqual(0, len(self.manager._index))
+
+    def test_callback_priority(self):
+        pri_first = priority_group.PRIORITY_DEFAULT - 100
+        pri_last = priority_group.PRIORITY_DEFAULT + 100
+        # lowest priority value should be first in the _callbacks
+        self.manager.subscribe(callback_1, 'my-resource', 'my-event')
+        self.manager.subscribe(callback_2, 'my-resource',
+                               'my-event', pri_last)
+        self.manager.subscribe(callback_3, 'my-resource',
+                               'my-event', pri_first)
+        callbacks = self.manager._callbacks['my-resource']['my-event']
+        # callbacks should be sorted based on priority for resource and event
+        self.assertEqual(3, len(callbacks))
+        self.assertEqual(pri_first, callbacks[0][0])
+        self.assertEqual(priority_group.PRIORITY_DEFAULT, callbacks[1][0])
+        self.assertEqual(pri_last, callbacks[2][0])
+
+    @mock.patch('neutron_lib.callbacks.manager.CallbacksManager._del_callback')
+    def test_del_callback_called_on_unsubscribe(self, mock_cb):
+        self.manager.subscribe(callback_1, 'my-resource', 'my-event')
+        callback_id = self.manager._find(callback_1)
+        callbacks = self.manager._callbacks['my-resource']['my-event']
+        self.assertEqual(1, len(callbacks))
+        self.manager.unsubscribe(callback_1, 'my-resource', 'my-event')
+        mock_cb.assert_called_once_with(callbacks, callback_id)
+
+    @mock.patch("neutron_lib.callbacks.manager.LOG")
+    def test_callback_order(self, _logger):
+        self.manager.subscribe(callback_1, 'my-resource', 'my-event', PRI_MED)
+        self.manager.subscribe(callback_2, 'my-resource', 'my-event', PRI_HIGH)
+        self.manager.subscribe(callback_3, 'my-resource', 'my-event', PRI_LOW)
+        self.assertEqual(
+            3, len(self.manager._callbacks['my-resource']['my-event']))
+        self.manager.unsubscribe(callback_3, 'my-resource', 'my-event')
+        self.manager.notify('my-resource', 'my-event', mock.ANY)
+        # callback_3 should be deleted and not executed
+        self.assertEqual(
+            2, len(self.manager._callbacks['my-resource']['my-event']))
+        self.assertEqual(0, callback_3.counter)
+        # executed callbacks should have counter incremented
+        self.assertEqual(1, callback_2.counter)
+        self.assertEqual(1, callback_1.counter)
+        callback_ids = _logger.debug.mock_calls[4][1][1]
+        # callback_2 should be first in exceution as it has higher priority
+        self.assertEqual(callback_id_2, callback_ids[0])
+        self.assertEqual(callback_id_1, callback_ids[1])
+
+    @mock.patch("neutron_lib.callbacks.manager.LOG")
+    def test__notify_loop_skip_log_errors(self, _logger):
+        self.manager.subscribe(
+            callback_raise, resources.PORT, events.BEFORE_CREATE)
+        self.manager.subscribe(
+            callback_raise, resources.PORT, events.PRECOMMIT_CREATE)
+        self.manager._notify_loop(
+            resources.PORT, events.BEFORE_CREATE, mock.ANY)
+        self.manager._notify_loop(
+            resources.PORT, events.PRECOMMIT_CREATE, mock.ANY)
+        self.assertFalse(_logger.exception.call_count)
+        self.assertTrue(_logger.debug.call_count)
+
+    def test_object_instances_as_subscribers(self):
+        """Ensures that the manager doesn't think these are equivalent."""
+        a = GloriousObjectWithCallback()
+        b = ObjectWithCallback()
+        c = ObjectWithCallback()
+        for o in (a, b, c):
+            self.manager.subscribe(
+                o.callback, resources.PORT, events.BEFORE_CREATE)
+            # ensure idempotency remains for a single object
+            self.manager.subscribe(
+                o.callback, resources.PORT, events.BEFORE_CREATE)
+        self.manager.notify(resources.PORT, events.BEFORE_CREATE, mock.ANY)
+        self.assertEqual(1, a.counter)
+        self.assertEqual(1, b.counter)
+        self.assertEqual(1, c.counter)
+
+    def test_publish_invalid_payload(self):
+        self.assertRaises(exceptions.Invalid, self.manager.publish,
+                          resources.PORT, events.AFTER_DELETE, self,
+                          payload=object())
+
+    def test_publish_empty_payload(self):
+        notify_payload = []
+
+        def _memo(resource, event, trigger, payload=None):
+            notify_payload.append(payload)
+
+        self.manager.subscribe(_memo, 'x', 'y')
+        self.manager.publish('x', 'y', self)
+        self.assertIsNone(notify_payload[0])
+
+    def test_publish_payload(self):
+        notify_payload = []
+
+        def _memo(resource, event, trigger, payload=None):
+            notify_payload.append(payload)
+
+        self.manager.subscribe(_memo, 'x', 'y')
+        payload = events.EventPayload(object())
+        self.manager.publish('x', 'y', self, payload=payload)
+        self.assertEqual(payload, notify_payload[0])
